@@ -10,27 +10,89 @@ import operator
 import sys
 import time
 from datetime import datetime
+from pathlib import Path
 from typing import Any, Callable, Dict, Mapping, Optional, Union
 
 import torch
+from azureml.core import Run, Workspace
 from pytorch_lightning import LightningModule, Trainer
 from pytorch_lightning.callbacks import ProgressBarBase
 from pytorch_lightning.loggers import LightningLoggerBase
 from pytorch_lightning.utilities.distributed import rank_zero_only
+from pytorch_lightning.utilities.logger import _convert_params, _flatten_dict, _sanitize_params
 
 from health_azure import is_running_in_azure_ml
-from health_azure.utils import RUN_CONTEXT
+from health_azure.utils import PathOrString, RUN_CONTEXT, create_aml_run_object
 
 
 class AzureMLLogger(LightningLoggerBase):
     """
-    A Pytorch Lightning logger that stores metrics in the current AzureML run. If the present run is not
-    inside AzureML, nothing gets logged.
+    A Pytorch Lightning logger that stores metrics in the current AzureML run. This logger will always write metrics
+    to AzureML if the training run is executed in AzureML. It can optionally also write to AzureML if the training
+    run is executed somewhere else, for example on a VM outside of AzureML.
     """
 
-    def __init__(self) -> None:
+    HYPERPARAMS_NAME = "hyperparams"
+    """
+    The name under which hyperparameters are written to the AzureML run.
+    """
+
+    def __init__(self,
+                 enable_logging_outside_azure_ml: Optional[bool] = False,
+                 experiment_name: str = "azureml_logger",
+                 run: Optional[Run] = None,
+                 run_name: Optional[str] = None,
+                 workspace: Optional[Workspace] = None,
+                 workspace_config_path: Optional[Path] = None,
+                 snapshot_directory: Optional[PathOrString] = None
+                 ) -> None:
+        """
+        :param enable_logging_outside_azure_ml: If True, the AzureML logger will write metrics to AzureML even if
+            executed outside of an AzureML run (for example, when working on a separate virtual machine). If False,
+            the logger will only write metrics to AzureML if the code is actually running inside of AzureML. Default
+            False, do not log outside of AzureML.
+        :param experiment_name: The AzureML experiment that should hold the run when executed outside of AzureML.
+        :param run: The AzureML run to log to when the ``enable_logging_outside_azure_ml`` flag is True. If None,
+            a new run will be created. When finished, the run should be completed by calling ``run.complete()``. The
+            logger itself only calls ``run.flush()`` in its ``finalize()`` method.
+        :param run_name: An optional name for the run (this will be used as the display name in the AzureML UI). This
+            argument only matters when running outside of AzureML.
+        :param workspace: If provided, use this workspace to create the run in.
+        :param workspace_config_path: Use this path to read workspace configuration json file. If not provided,
+            use the workspace specified by the `config.json` file in the current working directory or its parents.
+        :param snapshot_directory: The folder that should be included as the code snapshot. By default, no snapshot
+            is created. Set this to the folder that contains all the code your experiment uses. You can use a file
+            .amlignore to skip specific files or folders, akin to .gitignore..
+        """
         super().__init__()
         self.is_running_in_azure_ml = is_running_in_azure_ml()
+        self.run: Optional[Run] = None
+        self.has_user_provided_run = False
+        self.enable_logging_outside_azure_ml = enable_logging_outside_azure_ml
+        if self.is_running_in_azure_ml:
+            self.run = RUN_CONTEXT
+        elif enable_logging_outside_azure_ml:
+            if run is not None:
+                self.run = run
+                self.has_user_provided_run = True
+            else:
+                try:
+                    self.run = create_aml_run_object(experiment_name=experiment_name,
+                                                     run_name=run_name,
+                                                     workspace=workspace,
+                                                     workspace_config_path=workspace_config_path,
+                                                     snapshot_directory=snapshot_directory)
+                    # Display name should already be set when creating the run object, but this does not happen.
+                    # In unit tests, the run has the expected display name, but not here. Hence, set it again.
+                    self.run.display_name = run_name
+                except Exception as ex:
+                    logging.error(f"Unable to create an AzureML run to store the results because of {ex}.")
+                    raise
+            print(f"Writing metrics to run {self.run.id} in experiment {self.run.experiment.name}.")
+            print(f"To check progress, visit this URL: {self.run.get_portal_url()}")
+        else:
+            print("AzureMLLogger will not write any logs because it is running outside AzureML, and the "
+                  "'enable_logging_outside_azure_ml' flag is set to False")
 
     @rank_zero_only
     def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
@@ -43,33 +105,30 @@ class AzureMLLogger(LightningLoggerBase):
         :param step: The trainer global step for logging.
         """
         logging.debug(f"AzureMLLogger step={step}: {metrics}")
+        if self.run is None:
+            return
         is_epoch_metric = "epoch" in metrics
-        if self.is_running_in_azure_ml:
-            for key, value in metrics.items():
-                # Log all epoch-level metrics without the step information
-                # All step-level metrics with step
-                RUN_CONTEXT.log(key, value, step=None if is_epoch_metric else step)
+        for key, value in metrics.items():
+            # Log all epoch-level metrics without the step information
+            # All step-level metrics with step
+            self.run.log(key, value, step=None if is_epoch_metric else step)
 
     @rank_zero_only
     def log_hyperparams(self, params: Union[argparse.Namespace, Dict[str, Any]]) -> None:
         """
         Logs the given model hyperparameters to AzureML as a table. Namespaces are converted to dictionaries.
-        Nested dictionaries are flattened out.
+        Nested dictionaries are flattened out. The hyperparameters are then written as a table with two columns
+        "name" and "value".
         """
-        if not self.is_running_in_azure_ml:
+        if self.run is None:
             return
         if params is None:
             return
-        # Convert from Namespace to dictionary
-        params = self._convert_params(params)
-        # Convert nested dictionaries to folder-like structure
-        params = self._flatten_dict(params)
-        # Convert anything that is not a primitive type to str
-        params = self._sanitize_params(params)
-        if not isinstance(params, dict):
-            raise ValueError(f"Expected the hyperparameters to be a dictionary, but got {type(params)}")
-        if len(params) > 0:
-            RUN_CONTEXT.log_table("hyperparams", params)
+        params_final = _preprocess_hyperparams(params)
+        if len(params_final) > 0:
+            # Log hyperparameters as a table with 2 columns. Each "step" is one hyperparameter
+            self.run.log_table(self.HYPERPARAMS_NAME, {"name": list(params_final.keys()),
+                                                       "value": list(params_final.values())})
 
     def experiment(self) -> Any:
         return None
@@ -80,6 +139,17 @@ class AzureMLLogger(LightningLoggerBase):
     def version(self) -> int:
         return 0
 
+    def finalize(self, status: str) -> None:
+        if self.enable_logging_outside_azure_ml and not self.is_running_in_azure_ml and self.run is not None:
+            if self.has_user_provided_run:
+                # The logger uses a run that was provided by the user: Flush it, but do not complete it.
+                # The user should complete the run after finishing the experiment. This is important when running
+                # training outside of AzureML, so that training and inference metrics can be written to the same run.
+                self.run.flush()
+            else:
+                # Run.complete should only be called if we created an AzureML run here in the constructor.
+                self.run.complete()
+
 
 class AzureMLProgressBar(ProgressBarBase):
     """
@@ -119,7 +189,7 @@ class AzureMLProgressBar(ProgressBarBase):
         self._enabled = True
         self.stage = ""
         self.stage_start_time = 0.0
-        self.total_num_batches = 0
+        self.total_num_batches = 0.
         self.write_to_logging_info = write_to_logging_info
         self.print_timestamp = print_timestamp
 
@@ -135,6 +205,16 @@ class AzureMLProgressBar(ProgressBarBase):
     def is_disabled(self) -> bool:
         return not self.is_enabled
 
+    @property
+    def total_test_batches(self) -> int:
+        assert self._trainer is not None
+        return sum(self.trainer.num_test_batches)  # type: ignore
+
+    @property
+    def total_predict_batches(self) -> int:
+        assert self._trainer is not None
+        return sum(self.trainer.num_predict_batches)  # type: ignore
+
     def disable(self) -> None:
         self._enabled = False
 
@@ -157,7 +237,7 @@ class AzureMLProgressBar(ProgressBarBase):
         super().on_predict_epoch_start(trainer, pl_module)
         self.start_stage(self.PROGRESS_STAGE_PREDICT, self.total_predict_batches)
 
-    def start_stage(self, stage: str, total_num_batches: int) -> None:
+    def start_stage(self, stage: str, total_num_batches: Union[int, float]) -> None:
         """
         Sets the information that a new stage of the PL loop is starting. The stage will be available in
         self.stage, total_num_batches in self.total_num_batches. The time when this method was called is recorded in
@@ -204,6 +284,7 @@ class AzureMLProgressBar(ProgressBarBase):
         if not should_update:
             return
         prefix = f"{self.stage}"
+        assert self.trainer is not None and self.trainer.lightning_module is not None  # for pyright
         if self.stage in [self.PROGRESS_STAGE_TRAIN, self.PROGRESS_STAGE_VAL]:
             prefix += f" epoch {self.trainer.current_epoch}"
         if self.stage == self.PROGRESS_STAGE_TRAIN:
@@ -234,18 +315,36 @@ class AzureMLProgressBar(ProgressBarBase):
             sys.stdout.flush()
 
 
+def _preprocess_hyperparams(params: Any) -> Dict[str, str]:
+    """
+    Converts arbitrary hyperparameters to a simple dictionary structure, in particular argparse Namespaces.
+    Nested dictionaries are converted to folder-like strings, like ``{'a': {'b': 'c'}} -> {'a/b': 'c'}``.
+    All hyperparameter values are converted to strings, because Run.log_table can't deal with mixed datatypes.
+    :param params: The parameters to convert
+    :return: A dictionary mapping from string to string.
+    """
+    # Convert from Namespace to dictionary
+    params = _convert_params(params)
+    # Convert nested dictionaries to folder-like structure
+    params = _flatten_dict(params)
+    # Convert anything that is not a primitive type to str
+    params_final = _sanitize_params(params)
+    if not isinstance(params_final, dict):
+        raise ValueError(f"Expected the converted hyperparameters to be a dictionary, but got {type(params)}")
+    return {str(key): str(value) for key, value in params_final.items()}
+
+
 def log_on_epoch(module: LightningModule,
                  name: Optional[str] = None,
                  value: Optional[Any] = None,
                  metrics: Optional[Mapping[str, Any]] = None,
                  reduce_fx: Callable = torch.mean,
-                 sync_dist: Optional[bool] = None,
-                 sync_dist_op: Any = "mean") -> None:
+                 sync_dist: Optional[bool] = None) -> None:
     """
     Write a dictionary with metrics and/or an individual metric as a name/value pair to the loggers of the given module.
     Metrics are always logged upon epoch completion.
-    The metrics in question first synchronized across GPUs if DDP with >1 node is used, using the sync_dist_op
-    (default: mean). Afterwards, they are aggregated across all steps via the reduce_fx (default: mean).
+    The metrics in question first synchronized across GPUs if DDP with >1 node is used,. Afterwards, they are
+    aggregated across all steps via the reduce_fx (default: mean).
     Metrics that are fed in as plain numbers rather than tensors (for example, plain Python integers) are converted
     to tensors before logging, to enable synchronization across GPUs if needed.
 
@@ -258,8 +357,6 @@ def log_on_epoch(module: LightningModule,
         available on Rank 0 of a DDP job.
     :param reduce_fx: The reduce function to apply to the per-step values, after synchronizing the tensors across GPUs.
         Default: torch.mean
-    :param sync_dist_op: The reduce operation to use when synchronizing the tensors across GPUs. This must be
-        a value recognized by sync_ddp: 'sum', 'mean', 'avg'
     """
     assert module.trainer is not None, "No trainer is set for this module."
     if operator.xor(name is None, value is None):
@@ -278,8 +375,7 @@ def log_on_epoch(module: LightningModule,
                     on_epoch=True,
                     on_step=False,
                     sync_dist=is_sync_dist,
-                    reduce_fx=reduce_fx,
-                    sync_dist_op=sync_dist_op)
+                    reduce_fx=reduce_fx)
 
 
 def log_learning_rate(module: LightningModule, name: str = "learning_rate") -> None:
@@ -297,11 +393,11 @@ def log_learning_rate(module: LightningModule, name: str = "learning_rate") -> N
         raise ValueError("Learning rate logging can only be used during training.")
     single_scheduler = not isinstance(schedulers, list)
     if single_scheduler:
-        schedulers = [schedulers]
+        schedulers = [schedulers]  # type: ignore
     lr_0 = schedulers[0].get_last_lr()  # type: ignore
     singleton_lr = single_scheduler and len(lr_0) == 1
     logged = {}
-    for i, scheduler in enumerate(schedulers):
+    for i, scheduler in enumerate(schedulers):  # type: ignore
         for j, lr_j in enumerate(scheduler.get_last_lr()):  # type: ignore
             full_name = name if singleton_lr else f"{name}/{i}/{j}"
             logged[full_name] = lr_j
